/*
  Osu Keypad Code by Xieshi Zhang aka CrazyBlackStone
  Uses STM32F103C8T6 on custom PCB and Arduino IDE
  Project published under CC-BY-NC-SA
  Code under GPL v3 https://www.gnu.org/licenses/gpl-3.0.en.html
*/

#include <USBComposite.h>
#include <EEPROM.h> //eeprom library is natively installed on stm32duino or stm32 core

const char ManufacturerName[] = "Xieshi Zhang";
const char DeviceName[] = "Custom Osu! Keypad";
const char DeviceSerial[] = "00000000000000000001";

//pin definitions
#define red PA1
#define green PA2
#define blue PA3
#define switch1 PA4
#define switch2 PA5
#define switch3 PA6
#define switch4 PA7
#define modeSw PB0

//key definitions
char keys[4] = {'a', 's', 'z', 'x'};

//settings
#define debounceTime 5 //This is the debounce time in milliseconds for the keyboard switches. Changing this will not affect the latency but will determine the shortest keypress possible. Cherry specifies a 5ms maximum bounce time. 
#define maxPWM 255 //This is the maximum PWM value. It should be 255 by default
 #define RGBLED //This is the option for RGB LEDs that can be soldered on to the PCB. Comment this out if you're not using any LEDs. CAUTION: ALTHOUGH RGB LEDS DO NOT WASTE SIGNIFICANT CPU CYCLES, PLEASE DISABLE THIS OPTION IF NO LEDS ARE USED FOR MAXIMUM SPEED!
#define serialCommunication // This is the option for enabling/disabling composite serial, which lets you configure keys on the fly with the driver or with serial commands. WARNING: ALTHOUGH STM32F1 IS VERY FAST, LEAVING SERIAL INITIALIZED MAY WASTE CPU CYCLES!
#define ledDelay 2 //This is the LED color switching delay/speed.
#define colorIncrementDelay 5000 //This is the amount of time the color will stay before incrementing.

//variables
const int switchPins[4] = {switch1, switch2, switch3, switch4};
const int ledPins[3] = {red, green, blue};
bool oldSwitchState[4] = {LOW, LOW, LOW, LOW};
bool switchState[4] = {LOW, LOW, LOW, LOW};
bool oldModeState = HIGH;
unsigned long debounce[4] = {0, 0, 0, 0};
unsigned long modePressTime;
unsigned long ledTime;
unsigned long colorTime;
unsigned long currentMillis;
byte lightMode = 0;
byte rgbMode = 0;
byte colorCount;
int brightness = 50; //percent
float rgbBrightness[3] = {1, 1, 1};
float targetRGBBrightness[3] = {1, 1, 1};
byte pwmRGB[3] = {maxPWM * rgbBrightness[1] * brightness / 100, maxPWM * rgbBrightness[2] * brightness / 100, maxPWM * rgbBrightness[3] * brightness / 100};
String prefix = "i";

USBHID HID;
HIDKeyboard Keyboard(HID);
#ifdef serialCommunication
USBCompositeSerial CompositeSerial;
#endif

void setup() {
  pinMode(modeSw, INPUT_PULLUP);

  pinMode(switch1, INPUT);
  pinMode(switch2, INPUT);
  pinMode(switch3, INPUT);
  pinMode(switch4, INPUT);

  #ifdef RGBLED
  pinMode(red, OUTPUT);
  pinMode(green, OUTPUT);
  pinMode(blue, OUTPUT);

  if (EEPROM.read(7) != 1)
  {
    EEPROM.write(0, rgbMode);
    EEPROM.write(1, lightMode);
    EEPROM.write(2, brightness);
    EEPROM.write(7, 1);
  }
  
  rgbMode = EEPROM.read(0);
  lightMode = EEPROM.read(1);
  brightness = EEPROM.read(2);
  #endif
  
  USBComposite.setManufacturerString(ManufacturerName);
  USBComposite.setProductString(DeviceName);
  USBComposite.setSerialString(DeviceSerial);
  USBComposite.setVendorId(0x987);

  #ifdef serialCommunication
  HID.begin(CompositeSerial, HID_KEYBOARD);

  if (EEPROM.read(8) != 1)
  {
    EEPROM.write(3, keys[0]);
    EEPROM.write(4, keys[1]);
    EEPROM.write(5, keys[2]);
    EEPROM.write(6, keys[3]);
    EEPROM.write(8, 1);
  }
  
  keys[0] = EEPROM.read(3);
  keys[1] = EEPROM.read(4);
  keys[2] = EEPROM.read(5);
  keys[3] = EEPROM.read(6);
  #else
  HID.begin(HID_KEYBOARD);
  #endif
  
  Keyboard.begin();

  #ifdef RGBLED
  digitalWrite(red, HIGH);
  delay(400);
  digitalWrite(red, LOW);
  digitalWrite(green, HIGH);
  delay(400);
  digitalWrite(green, LOW);
  digitalWrite(blue, HIGH);
  delay(400);
  digitalWrite(blue, LOW);

  checkRGBMode();
  if (lightMode == 0)
    turnOnLED();
  else
    turnOffLED();
  #endif
}

void loop() {
  //key polling and debouncing
  for (int i = 0; i < 4; i++)
  {
    switchState[i] = digitalRead(switchPins[i]);
    if (switchState[i] && !oldSwitchState[i] && (millis() - debounce[i]) > debounceTime)
    {
      debounce[i] = millis();
      oldSwitchState[i] = switchState[i];
      Keyboard.press(keys[i]);

      #ifdef RGBLED
      if (lightMode == 2)
        turnOnLED();
      #endif
    }
    else if (!switchState[i] && oldSwitchState[i] && (millis() - debounce[i]) > debounceTime)
    {
      debounce[i] = millis();
      oldSwitchState[i] = switchState[i];
      Keyboard.release(keys[i]);

      #ifdef RGBLED
      if (lightMode == 2)
        turnOffLED();
      #endif
    }
  }

  #ifdef serialCommunication
  if (CompositeSerial.available())
  {
    String command = CompositeSerial.readString();
    if (command.startsWith("i"))
    {
      prefix = "i";
      printInfo();
    }
    else if (command.startsWith("s")) //s:1:k:
    {
      String setKey = getValue(command, ':', 1);
      String setString = getValue(command, ':', 2);
      int intSetKey = setKey.toInt();
      if (setString[0] != ':')
        keys[intSetKey - 1] = setString[0];
      else
        keys[intSetKey - 1] = NULL;
      int EPValue = intSetKey + 2;
      EEPROM.update(EPValue, setString[0]);
      prefix = "d";
      printInfo();
    }
  }
  #endif
  //rgb led control, will be disabled if not defined
  #ifdef RGBLED
  if (digitalRead(modeSw) != oldModeState)
  {
    delay(10);
    bool modeState = digitalRead(modeSw);
    if (modeState == LOW && modeState != oldModeState)
    {
      modePressTime = millis();
      oldModeState = modeState;
    }
    else if (modeState == HIGH && modeState != oldModeState)
    {
      oldModeState = modeState;
      if ((millis() - modePressTime) < 300)
      {
        if (lightMode < 2)
          lightMode++;
        else
          lightMode = 0;
        EEPROM.update(1, lightMode);
      }
      else if ((millis() - modePressTime) <= 4000 && lightMode != 1)
      {
        if (rgbMode < 7)
          rgbMode++;
        else
          rgbMode = 0;
        EEPROM.update(0, rgbMode);
        checkRGBMode();
      }

      if (lightMode == 0)
        turnOnLED();
      else
        turnOffLED();

      if (lightMode == 1)
      {
        rgbBrightness[0] = 0;
        rgbBrightness[1] = 0;
        rgbBrightness[2] = 0;
      }
      //turnOnLED();
    }
  }
  else if (!digitalRead(modeSw) && (millis() - modePressTime) > 4000 && modePressTime)
  {
    currentMillis = millis();
    if ((currentMillis - modePressTime) < 5000)
    {
      brightness = 100;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 5500)
    {
      brightness = 90;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 6000)
    {
      brightness = 80;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 6500)
    {
      brightness = 70;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 7000)
    {
      brightness = 60;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 7500)
    {
      brightness = 50;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 8000)
    {
      brightness = 40;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 8500)
    {
      brightness = 30;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 9000)
    {
      brightness = 20;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 9500)
    {
      brightness = 10;
      turnOnLED();
    }
    else if ((currentMillis - modePressTime) < 10000)
    {
      brightness = 5;
      turnOnLED();
    }
    else
    {
      brightness = 0;
      turnOnLED();
    }
    EEPROM.update(2, brightness);
  }

  if (lightMode == 1)
  {
    if ((millis() - ledTime) >= ledDelay)
    {
      ledTime = millis();
      for (int i = 0; i < 3; i++)
      {
        if (rgbBrightness[i] > targetRGBBrightness[i])
          rgbBrightness[i] = rgbBrightness[i] - 0.01;
        else if (rgbBrightness[i] < targetRGBBrightness[i])
          rgbBrightness[i] = rgbBrightness[i] + 0.01;
        if (rgbBrightness[i] > 1)
          rgbBrightness[i] = 1;
      }
      turnOnLED();
    }
    if ((millis() - colorTime) >= colorIncrementDelay)
    {
      colorTime = millis();
      incrementColor();
    }
  }
#endif
}

#ifdef RGBLED
void turnOnLED()
{
  for (int i = 0; i < 3; i++)
  {
    pwmRGB[i] = maxPWM * rgbBrightness[i] * brightness / 100;
    analogWrite(ledPins[i], pwmRGB[i]);
  }
}
void turnOffLED()
{
  for (int i = 0; i < 3; i++)
  {
    analogWrite(ledPins[i], 0);
  }
}
void checkRGBMode()
{
  if (rgbMode == 0)
  {
    rgbBrightness[0] = 1;
    rgbBrightness[1] = 0;
    rgbBrightness[2] = 0;
  }
  else if (rgbMode == 1)
  {
    rgbBrightness[0] = 0.7;
    rgbBrightness[1] = 0.4;
    rgbBrightness[2] = 0;
  }
  else if (rgbMode == 2)
  {
    rgbBrightness[0] = 0;
    rgbBrightness[1] = 1;
    rgbBrightness[2] = 0;
  }
  else if (rgbMode == 3)
  {
    rgbBrightness[0] = 0;
    rgbBrightness[1] = 0.5;
    rgbBrightness[2] = 0.5;
  }
  else if (rgbMode == 4)
  {
    rgbBrightness[0] = 0;
    rgbBrightness[1] = 0;
    rgbBrightness[2] = 1;
  }
  else if (rgbMode == 5)
  {
    rgbBrightness[0] = 0.5;
    rgbBrightness[1] = 0;
    rgbBrightness[2] = 0.5;
  }
  else if (rgbMode == 6)
  {
    rgbBrightness[0] = 0.36;
    rgbBrightness[1] = 0.14;
    rgbBrightness[2] = 0.12;
  }
  else if (rgbMode == 7) //ffffe0
  {
    rgbBrightness[0] = 0.8;
    rgbBrightness[1] = 0.2;
    rgbBrightness[2] = 0.1;
  }
}
void incrementColor()
{
  if (colorCount < 1)
    colorCount++;
  else
    colorCount = 0;

  if (colorCount == 0)
  {
    targetRGBBrightness[0] = 1;
    targetRGBBrightness[1] = 0;
    targetRGBBrightness[2] = 0;
  }
  else if (colorCount == 1)
  {
    targetRGBBrightness[0] = 0;
    targetRGBBrightness[1] = 0;
    targetRGBBrightness[2] = 1;
  }
}
#endif

#ifdef serialCommunication
String getValue(String data, char separator, int index)
{
    int found = 0;
    int strIndex[] = { 0, -1 };
    int maxIndex = data.length() - 1;

    for (int i = 0; i <= maxIndex && found <= index; i++) {
        if (data.charAt(i) == separator || i == maxIndex) {
            found++;
            strIndex[0] = strIndex[1] + 1;
            strIndex[1] = (i == maxIndex) ? i+1 : i;
        }
    }
    return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}
void printInfo()
{
  CompositeSerial.print(prefix);
  CompositeSerial.print(":");
  if (keys[0] != ':')
    CompositeSerial.print(keys[0]);
  CompositeSerial.print(":");
  if (keys[1] != ':')
    CompositeSerial.print(keys[1]);
  CompositeSerial.print(":");
  if (keys[2] != ':')
    CompositeSerial.print(keys[2]);
  CompositeSerial.print(":");
  if (keys[3] != ':')
    CompositeSerial.print(keys[3]);
  CompositeSerial.print(":");
  CompositeSerial.print("v1.0.0");
  #ifdef RGBLED
  CompositeSerial.print(":");
  CompositeSerial.print(lightMode);
  CompositeSerial.print(":");
  CompositeSerial.print(rgbMode);
  CompositeSerial.print(":");
  CompositeSerial.print(brightness);
  #endif
  CompositeSerial.print("\n");
}
#endif
